/* Radix-2 Fast Fourier Transform, first shot
 * Demonstration code by Emil Mikulic, 2001.
 * http://dmr.ath.cx/sound/fft/
 *
 * Feel free to do whatever you like with this code.
 * Feel free to credit me.
 */

#include <math.h>
#include <stdio.h>
#include <sndfile.h> /* http://www.zip.com.au/~erikd/libsndfile/ */

typedef double val;
#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
/* overkill */
#  define PI 3.141592653589793238462643383279502884197169399375105820974
# endif
#endif

val *in_re, *in_im, *out_re, *out_im;
SF_INFO sfinfo;
SNDFILE *sndfile;

/* note: N must be a power of 2 or whacked-out things will happen */
void fft(val *in_re, val *in_im, int N, val *out_re, val *out_im)
{
	val *fo_re, *fe_re, *fo_im, *fe_im, *tmp_re, *tmp_im;
	int i, ofs;

	/* trivial */
	if (N == 1)
	{
		out_re[0] = in_re[0];
		out_im[0] = in_im[0];
		return;
	}

#define CHECK(x) if ((x) == NULL)\
	{ fprintf(stderr,"malloc failed!\n"); exit(1); }

	CHECK( fo_re = (val*)malloc(sizeof(val) * N/2) );
	CHECK( fo_im = (val*)malloc(sizeof(val) * N/2) );
	CHECK( fe_re = (val*)malloc(sizeof(val) * N/2) );
	CHECK( fe_im = (val*)malloc(sizeof(val) * N/2) );

	CHECK( tmp_re = (val*)malloc(sizeof(val) * N) );
	CHECK( tmp_im = (val*)malloc(sizeof(val) * N) );

#undef CHECK

	/* split into even, odd */
	for (i=0; i<N/2; i++)
	{
		fe_re[i] = in_re[i*2];
		fe_im[i] = in_im[i*2];

		fo_re[i] = in_re[i*2+1];
		fo_im[i] = in_im[i*2+1];
	}

	fft(fe_re, fe_im, N/2,   tmp_re,         tmp_im);
	fft(fo_re, fo_im, N/2, &(tmp_re[N/2]), &(tmp_im[N/2]));

	/* fudge odds */
	for (i=0; i<N/2; i++)
	{
		val t_re, t_im, theta, or, oi;

		theta = -2.0 * PI * (val)i / (val)N;
		t_re = cos(theta);
		t_im = sin(theta);

		or = tmp_re[i+N/2]*t_re - tmp_im[i+N/2]*t_im;
		oi = tmp_im[i+N/2]*t_re + tmp_re[i+N/2]*t_im;

		out_re[i] = tmp_re[i] + or;
		out_im[i] = tmp_im[i] + oi;

		out_re[i+N/2] = tmp_re[i] - or;
		out_im[i+N/2] = tmp_im[i] - oi;
	}

	free(fe_re);
	free(fe_im);
	free(fo_re);
	free(fo_im);
	free(tmp_re);
	free(tmp_im);
}



int main()
{
	int i, N;

	sndfile = sf_open_read("sample.wav", &sfinfo);

	/* Informational */
	fprintf(stderr, "Opened sample --\n\n"
		"samplerate:  %d\n"
		"samples:     %d\n"
		"channels:    %d\n"
		"pcmbitwidth: %d\n\n",
		sfinfo.samplerate, sfinfo.samples,
		sfinfo.channels, sfinfo.pcmbitwidth);

	if (sfinfo.channels != 1) {
		fprintf(stderr, "Need mono input.\n");
		sf_close(sndfile);
		return 1;
	}

	/* Allocate memory */
	in_re = (val*)malloc(sizeof(val) * sfinfo.samples);
	in_im = (val*)malloc(sizeof(val) * sfinfo.samples);
	out_re = (val*)malloc(sizeof(val) * sfinfo.samples);
	out_im = (val*)malloc(sizeof(val) * sfinfo.samples);

	if (!in_re || !in_im || !out_re || !out_im) {
		fprintf(stderr, "Out of memory!\n");
		sf_close(sndfile);
		return 1;
	}

	/* Fill up input */
	if (sf_read_double(sndfile, in_re, sfinfo.samples, 1) !=
		sfinfo.samples) {
		fprintf(stderr, "Failed to read %d samples.\n",
			sfinfo.samples);
		sf_close(sndfile);
		return 1;
	}
	sf_close(sndfile);

	/* Zeros */
	for (i=0; i<sfinfo.samples; i++)
		in_im[i] = 0;

	N = 65536;
	fft(in_re, in_im, N, out_re, out_im);

	/* Output */
	for (i=0; i<N/2; i++)
	{
		out_re[i] /= (val)N;
		out_im[i] /= (val)N;

		printf("%f Hz: %f\n",
			(double)i*(val)sfinfo.samplerate/(val)N,
			sqrt(out_re[i]*out_re[i] + out_im[i]*out_im[i]) );
	}

	free(in_im); free(in_re);
	free(out_im); free(out_re);
	return 0;
}
